poly.fit <- function(data, cov, weight, penalty, maxdim, exclude, include,
   additive = F, linear, delete=2, fit, silent = T, normweight = T, tdata, 
   tcov, tweight, cv, select=0, loss, seed)
{
   call <- match.call()
   it <- 0
   ntdata <- 0
   if(!missing(cv)) it <- 2
   if(!missing(tdata))it <- 1
        if(missing(cv)) cv <- 0
        if(it==1||it==0) cv <- 0
   if(it==2){
      if(!missing(seed)){
         if(sum(seed)!=0){
            if(length(seed)>11).Random.seed <- seed[1:12]
            else set.seed(seed[1])
            seed <- .Random.seed
         }
      }
      else{
         if(!missing(fit)){
            if(fit$method==2) .Random.seed <- fit$seed
         }
         seed <- .Random.seed
      }
   }
   z <- .C("spoly", mk = as.integer(rep(-3,13)))
   MAXKNOTS <- z$mk[1]
   MAXSPACE <- z$mk[2]    
   if(missing(data)) stop("there has to be data")
   if(length(data) < 25) stop("not enough data")
   if(is.integer(data) == F){
      if(max(abs(as.integer(data) - data)) < 0.001)
         data <- as.integer(data)
      else stop("not-integer data")
   }
   if(it == 1) {
      if(is.integer(tdata) == F){
         if(max(abs(as.integer(tdata) - tdata)) < 0.001)
            tdata <- as.integer(tdata)
         else stop("not-integer test data")
      }
      alldata <- c(data,tdata)
      if(min(alldata)<0) stop("negative data")
      clss <- min(alldata):max(alldata)
      if(min(alldata) == 1){
         data <- data - 1
         tdata <- tdata - 1
      }
      ntdata <- length(tdata)
      if(missing(tweight)) tweight <- rep(1,ntdata)
      if(length(tweight)!=ntdata)stop("length tweight is incorrect")
      if(normweight == T)tweight <- tweight*ntdata/sum(tweight)
   }
   else{
      if(min(data)<0) stop("negative data")
      clss <- min(data):max(data)
      if(min(data) == 1) data <- data - 1
   }
   nclass <- length(clss)
   ndata <- length(data)   
   xx <- .Random.seed 
   yy <- sample(ndata)
   .Random.seed <- xx
   if(missing(weight)) weight <- rep(1,ndata)
   if(it==2){
      if(sum(abs(seed))==0) myord <- 1:ndata
      else myord <- sample(ndata)
      data <- data[myord]
      weight <- weight[myord]
   }
   if(length(weight)!=ndata)stop("length weight is incorrect")
   if(normweight == T)weight <- weight*ndata/sum(weight)
   if(missing(cov)) {
      stop("covariates required")
   }
   else {
      if(length(cov) == ndata)
         cov <- matrix(cov, ncol = 1, nrow = ndata)
      if(length(cov[, 1]) != ndata)
         stop("covariates not ndata * ncov matrix")
      if(it==2)cov <- cov[myord,]
      ncov <- length(cov[1,  ])
      nms <- 1:ncov
      if(is.matrix(cov))
         nms <- dimnames(cov)[[2]]
      if(length(nms) != ncov)
         nms <- 1:ncov
   }
   if(missing(penalty) && it ==0)
      penalty <- log(ndata)
   if(missing(penalty) && it >0)
      penalty <- 0
   il <- 1
   if(select==1) il <- 0
   if(select==2) il <- 2
   if(delete!=0 && delete !=1) delete <- 2
   if(missing(loss)) loss <- 1 - diag(rep(1,nclass))
   if(il!=1 && !missing(loss))
      stop("if loss is specified, select has to be 0")
   if((it == 0) && !missing(loss))
      warning("loss only has effect when there is a test-set or CV is used")
   if(it == 1){
      if(missing(tcov)) {
         if(ncov!=0)stop("missing tcov")
         tcov <- 0
      }
      else {
         if(length(tcov) == ntdata)
         tcov <- matrix(tcov, ncol = 1, nrow = ntdata)
         if(length(tcov[, 1]) != ntdata)
         stop("test-covariates not ntdata * ncov matrix")
         ntcov <- length(cov[1,  ])
         if(ntcov!=ncov) stop("wrong number of test-covariates")
      }
   }
   naction <- nclass
   if(it>0){
      if(is.matrix(loss)==F)stop("loss is not a matrix")
      if(length(loss[1,])!=nclass)stop("loss has not nclass columns")
      naction <- length(loss[,1])
   }
   if(additive) {
      if(!missing(exclude)) stop("cannot have exclude and additive")
      if(!missing(include)) stop("cannot have include and additive")
      include <- c(0, 0)
   }
   if(missing(exclude) + missing(include) == 0)
      stop("only 1 from exclude and include allowed")
   vexclude <- 0   
   if(missing(exclude) == F) {
      if(length(exclude) == 2)
         exclude <- matrix(exclude, ncol = 2, nrow = 1)
      if(length(exclude[1,  ]) != 2) stop("exclude has wrong shape")
      if(min(exclude) < 0 || max(exclude) > ncov)
         stop("exclude has wrong values")
      vexclude <- as.vector(t(exclude))
      vexclude <- c(length(vexclude)/2, vexclude)   
   }
   if(missing(include) == F || additive) {
      if(length(include) == 2)
         include <- matrix(include, ncol = 2, nrow = 1)
      if(length(include[1,  ]) != 2)
         stop("include has wrong shape")
      if(min(include) < 0 || max(include) > ncov)
         stop("include has wrong values")
      include <- t(apply(include, 1, sort))
      if(length(include) == 2)
         include <- matrix(include, ncol = 2, nrow = 1)
      vexclude <- as.vector(t(include))
      vexclude <- c( - length(vexclude)/2, vexclude)
   }
   if(missing(maxdim)) {
      maxdim <- floor(4 * (ndata)^(1/3))+1
      maxdim <- min(ndata/2, MAXSPACE-1, (nclass-1)*maxdim)
      maxdim <-  - maxdim
   }
   if(maxdim > MAXSPACE - 1) {
      maxdim <- MAXSPACE - 1
      print(paste("maximum dimension reduced to", maxdim))
   }
   lins <- rep(0, MAXSPACE)
   if(!missing(linear)) {
      linear[linear <= 0] <- ncov + 1
      linear[linear > ncov + 1] <- ncov + 1
      lins[linear] <- 1
   }
   if(additive)
      vexclude <- c(-1, 0, 0)   # do it
   fitter <- 0
   bbtt <- matrix(0, ncol = 4 + max(data), nrow = abs(maxdim))
   cckk <- matrix(0, ncol = (MAXKNOTS + 1), nrow = ncov+1)
   if(!missing(fit)) {
      fitter <- (fit$nclass-1)*(fit$nbas)
      if(fit$ncov != ncov)
         stop("ncov and fit's ncov are different")
      if(fit$nclass != nclass)
         stop("nclass and fit's nclass are different")
                a1 <- length(fit$fcts[1,])
      bbtt[1:fit$nbas,  ] <- fit$fcts[,-a1]
      bbtt <- as.vector(t(bbtt))
      bbtt[is.na(bbtt)] <- -1
      a1 <- length(fit$knots[1,  ])
                a2 <- as.vector(t(fit$knots))
      cckk <- as.vector(cckk)
                cckk <- c(a1,a2,cckk)
      cckk[is.na(cckk)] <- -1
   }
   mindist <- 3*nclass
        if(missing(tdata)){
           tdata<-0
      tcov <-0
           tweight <- 0
        }
   ranges <- NA
   if(ncov == 1)
      ranges <- matrix(range(cov), ncol = 1, nrow = 2)
   if(ncov > 1)
      ranges <- apply(cov, 2, range)   # done
   cov <- as.single(t(cov))
   aicx <- as.single(rep(0,1000))
   intpars <-c(ndata,nclass,ncov,mindist,maxdim,silent,fitter,cv,it,ntdata,
             naction,il,delete)
   anova <- loss
   if(length(anova)<abs(maxdim)*4)anova<-c(anova,rep(0,abs(4*maxdim)))
   z <- .C("spoly",
      intpars = as.integer(intpars),
      as.integer(data),
      as.single(cov),
      anova = as.double(anova),
      as.double(penalty),
      bbtt = as.double(bbtt),
      cckk = as.double(cckk),
      as.integer(vexclude),
      as.integer(lins),
      logl = as.double(rep(0, 11*MAXSPACE+1)),
      as.double(weight),
      as.integer(tdata),
      as.single(t(tcov)),
      as.double(tweight),
      bbb = as.double(rep(0, MAXSPACE*nclass)),
      aicx=as.single(aicx))
   ndim <- z$intpars[1]
   aicx <- z$aicx[1:4]
   aicy <- 0
   if(it==2){
      aicy <- z$aicx[6:(z$aicx[5])]
      aicy <- matrix(aicy,ncol=3,byrow=T)
      if(z$aicx[5]<995)aicy[length(aicy[,1]),2]<- Inf
      dimnames(aicy) <- list(NULL,c("pen-min","pen-max","cv-loss"))
   }
   nclass <- z$intpars[2]+1
   nbas <- z$intpars[3]
   maxdim <- abs(maxdim)
   z$bbtt <- matrix(z$bbtt, nrow = maxdim, ncol = 3 + nclass, byrow =  T)
   z$bbtt <- z$bbtt[1:nbas,  ]
   z$cckk <- matrix(z$cckk, nrow = ncov, ncol = MAXKNOTS + 1, byrow = T)
   z$cckk <- z$cckk[, 1:(1 + max(z$cckk[, 1]))]
   z$cckk <- matrix(z$cckk, nrow = ncov)
   l1 <- max(z$cckk[, 1])
   for(i in 1:(ncov))
      if(z$cckk[i, 1] != l1) z$cckk[i, (z$cckk[i, 1] + 2):(l1 + 1)] <- 
            NA
   if(l1 > 0)
      dimnames(z$cckk) <- list(nms, c("K", 1:l1))
   if(l1 == 0)
      dimnames(z$cckk) <- list(nms, "K")
   z$bbtt <- matrix(z$bbtt, ncol = 3 + nclass)
   z$bbtt <- cbind(z$bbtt, 0)
   dimnames(z$bbtt) <- list(1:nbas, c("dim1", "knot1", 
      "dim2", "knot2", as.character(clss)))
   z$bbtt[z$bbtt[, 3] == -1, 3:4] <- NA
   z$bbtt[z$bbtt[, 4] == 0, 4] <- NA
   z$bbtt[1,1] <- NA
   i <- z$logl[1]
   z$logl <- matrix(z$logl[2:(11*i+1)],ncol=11,byrow=T)
   z$logl[z$logl[,10]<0,10] <- NA
   z$logl[z$logl[,11]<0,11] <- NA
   z$logl[1,11] <- Inf
   dimnames(z$logl) <- list(NULL, c("dim","loss","l-lik-trn","loss-trn",
      "sq-err-trn","l-lik-test","loss-tst","sq-err-tst"
              , "A/D","pen-min","pen-max"))
   if(it!=1){
      dimnames(z$logl)[[2]][2] <- "AIC"
      z$logl <- z$logl[,-(6:8)]
   }
   anova <- z$anova[2:(1+z$anova[1])]
   anova[anova<0] <- NA
   anova <- matrix(anova,ncol=3,byrow=T)
   dimnames(ranges) <- list(c("min", "max"), nms)
   z$bbtt[0, 0] <- NA
   z$bbtt[z$bbtt[, 2] == 0, 2] <- NA
   z$bbtt[z$bbtt[, 2] == 0, 4] <- NA
   if(nclass==naction)
      yyy <- clss
   else
      yyy <- 1:naction
   if(it!=0)dimnames(loss) <- list(as.character(yyy),clss)
   if(il!=1)loss <- -1
   bbb <- z$bbb
   bbb <- bbb[1:(nbas*nclass)]
   bbb <- matrix(bbb,nrow=nbas,byrow=T)
   if(it==0){
      list(call = call, ncov = ncov, ndim = ndim, nclass = nclass,
      nbas = nbas, fcts = z$bbtt, knots = z$cckk, penalty = penalty,
      method = it, ranges = ranges, logl= z$logl, 
      sample = ndata, wgtsum = sum(weight), covnames = nms,
      classnames = clss, beta = bbb, delete = delete, anova = anova)
   }
   else{
      if(it==1)
         list(call = call, ncov = ncov, ndim = ndim, nclass = nclass,
         nbas = nbas, naction = naction, fcts = z$bbtt, knots = z$cckk,
         loss = loss, penalty = penalty, method = it, ranges = ranges,
         logl= z$logl, sample = ndata, tsample = ntdata,
         wgtsum = sum(weight), covnames = nms, classnames = clss, beta = bbb,
         delete = delete, anova = anova,
         select = select, twgtsum = sum(tweight))
      else
         list(call = call, ncov = ncov, ndim = ndim, nclass = nclass,
         nbas = nbas, naction = naction, fcts = z$bbtt, knots = z$cckk,
         cv = cv, loss = loss, penalty = penalty, method = it, ranges = ranges,
         logl= z$logl, sample = ndata, wgtsum = sum(weight), covnames = nms,
         classnames = clss, cv.aic = aicx, cv.tab = aicy, seed = seed, 
         beta = bbb, delete = delete, anova = anova, select = select)
   }
}
cpoly <- function(cov, fit)
{
   xxx <- ppoly(cov, fit)
   yyy <- fit$classnames
   if(fit$method!=0){
   if(length(fit$loss)==1)
          fit$loss <- 1 - diag(rep(1,fit$nclass))
   xxx <- t(-fit$loss%*%t(xxx))
   if(fit$nclass==fit$naction)
      yyy <- fit$classnames
   else
      yyy <- 1:fit$naction
   }
   zzz <- xxx[, 1]
   www <- rep(yyy[1], length(zzz))
   for(i in 2:length(yyy)) {
      www[zzz < xxx[, i]] <- yyy[i]
      zzz[zzz < xxx[, i]] <- xxx[zzz < xxx[, i], i]
   }
   www
}
ppoly <- function(data, cov, fit)
{
   if(missing(fit)) {
      fit <- cov
      cov <- data
   }
   if(is.matrix(cov) == F)
      cov <- matrix(cov, ncol = 1)
   if(length(cov[1,  ]) != fit$ncov) {
      if(length(cov[1,  ]) == 1 && length(cov[, 1]) == fit$ncov)
         cov <- t(cov)
      else stop("incorrect number of covariates")
   }
   ncase <- length(cov[, 1])
   nclass <- fit$nclass
   nbas <- length(fit$fcts[, 1])
   if(missing(fit) || missing(data))
      data <- rep(-1, ncase)
   if(length(data) == 1)
      data <- rep(data, ncase)
   if(is.integer(data) == F)
      if(max(abs(as.integer(data) - data)) < 0.001)
         data <- as.integer(data)
      else stop("not-integer data")
   w2 <- fit$classnames
   if(data[1] != -1 && (min(w2) > min(data) || max(w2) < max(data)))
      stop("data has wrong range")
   if(min(data) == 0)
      data <- data + 1
   ppp <- matrix(0, ncol = nclass, nrow = ncase)
   for(i in 1:(nclass - 1))
      ppp[, i] <- (fit$fcts[1, (4 + i)])
   if(nbas > 1)
      for(j in 2:nbas) {
         uuu <- cov[, fit$fcts[j, 1]]
         if(is.na(fit$fcts[j, 2]) == F) {
            uuu <- uuu - fit$knots[fit$fcts[j, 1], fit$fcts[
              j, 2] + 1]
            uuu[uuu < 0] <- 0
         }
         vvv <- rep(1, ncase)
         if(is.na(fit$fcts[j, 3]) == F) {
            vvv <- cov[, fit$fcts[j, 3]]
            if(is.na(fit$fcts[j, 4]) == F) {
              vvv <- vvv - fit$knots[fit$fcts[j, 3], fit$
                fcts[j, 4] + 1]
              vvv[vvv < 0] <- 0
            }
         }
         uuu <- uuu * vvv
         for(i in 1:(nclass - 1))
            ppp[, i] <- ppp[, i] + uuu * fit$fcts[j, (4 + i
              )]
      }
   ppp <- exp(ppp)
   zzz <- ppp[, nclass]
   for(i in 1:(nclass - 1))
      zzz <- zzz + ppp[, i]
   for(i in 1:nclass)
      ppp[, i] <- ppp[, i]/zzz
   if(data[1] == -1)
      dimnames(ppp) <- list(NULL, fit$classnames)
   else ppp <- ppp[cbind(1:ncase, data)]
   ppp
}
poly.plot <- function(cov, fit, which, lims, what, data, n, xlab, ylab, zlab, 
   ...)
{
   here <- c(-1, -1)
   if(length(which) == 1 || length(which) == 2)
      here[1] <- length(which)
   here[2] <- as.integer(what)
   if(here[2]< 1||here[2]>8) stop("what is wrong")
   if(min(here) < 0)
      stop("which is wrong")
   if(here[2] < 5.5 && here[1] == 1)
      stop("which and what contradict")
   if(here[2] > 5.5 && here[1] == 2)
      stop("which and what contradict")
   if(length(cov) != fit$ncov)
      stop("length of cov is wrong")
   clbs <- fit$covnames
   ww <- fit$classnames
   w1 <- (1:fit$ncov)
   if(missing(lims))
      lims <- NULL
   if(length(lims) != 0 && length(lims) != (here[1] * 2))
      stop("lims is wrong")
   wa_0
   for(i in 1:length(which)){
   if(is.numeric(which) == F)
      wa <- c(wa,w1[which[i] == clbs])
   else wa <- c(wa,w1[w1 == which[i]])
   }
   wa <- wa[-1]
   if(length(wa) != here[1])
      stop("which is wrong")
   wb <- clbs[wa]
   if(here[2] < 3.5 || here[2] > 7.5) {
      if(missing(data))
         stop("data is missing")
      if(length(data) > 1)
         stop("only one class (data) allowed")
   }
   if(length(lims) == 0) {
      if(here[1] == 1)
         lims <- fit$ranges[, wa]
      else lims <- c(fit$ranges[, wa[1]], fit$ranges[, wa[2]])
   }
   if(missing(xlab))
      xlab <- as.character(wb[1])
   if(missing(ylab)) {
      if(here[1] == 2)
         ylab <- as.character(wb[2])
      if(here[2] > 6.5)
         ylab <- "probability"
      if(here[2] == 6)
         ylab <- "class"
   }
   if(missing(zlab) && here[2] == 2)
      zlab <- "probability"
   if(missing(n) && here[1] == 1)
      n <- 250
   if(missing(n) && here[1] == 2)
      n <- 50
   if(here[1] == 1) {
      cov <- matrix(cov, byrow = T, nrow = n, ncol = fit$ncov)
      c1 <- lims[1] + ((lims[2] - lims[1]) * (0:(n - 1)))/(n - 1)
      cov[, wa] <- c1
   }
   if(here[1] == 2) {
      cov <- matrix(cov, byrow = T, nrow = n * n, ncol = fit$ncov)
      c1 <- lims[1] + ((lims[2] - lims[1]) * (0:(n - 1)))/(n - 1)
      c11 <- (rep(c1, n))
      cov[, wa[1]] <- c11
      c2 <- lims[3] + ((lims[4] - lims[3]) * (0:(n - 1)))/(n - 1)
      c22 <- sort(rep(c2, n))
      cov[, wa[2]] <- c22
   }
   if(here[2] <= 3) {
      v1 <- ppoly(data, cov, fit)
      v1 <- matrix(v1, n, n)
      if(here[2] == 1)
         contour(c1, c2, v1, xlab = xlab, ylab = ylab, ...)
      if(here[2] == 2)
         persp(c1, c2, v1, xlab = xlab, ylab = ylab, zlab = zlab,
            ...)
      if(here[2] == 3)
         image(c1, c2, v1, xlab = xlab, ylab = ylab, ...)
   }
   if(here[2] == 6) {
      v1 <- cpoly(cov, fit)
      plot(c1, v1, type = "l", ylim = range(ww), xlab = xlab, ylab = 
         ylab, ...)
   }
   if(here[2] == 8) {
      v1 <- ppoly(data, cov, fit)
      plot(c1, v1, type = "l", xlab = xlab, ylab = ylab, ...)
   }
   if(here[2] == 4 || here[2] == 5) {
      v1 <- cpoly(cov, fit)
      v1 <- matrix(v1, n, n)
      if(here[2] == 5)
         image(c1, c2, v1, xlab = xlab, ylab = ylab, ...)
   }
   if(here[2] == 4) {
      zz <- range(v1)
      z1 <- 1 * (v1 < zz[1] + 0.5)
      contour(c1, c2, z1, xlab = xlab, ylab = ylab, levels = 0.5, 
         labex = 0, ...)
      if(zz[2] - zz[1] > 1)
         for(i in (zz[1] + 1):(zz[2] - 1)) {
            z1 <- 1 * (v1 < i + 0.5)
            contour(c1, c2, z1, labex = 0, levels = 0.5, 
              add = T, ...)
         }
   }
   if(here[2] == 7) {
      v1 <- ppoly(cov, fit)
      plot(c1, v1[, 1], type = "l", xlab = xlab, ylab = ylab, ylim = 
         c(0, 1), ...)
      abline(h = c(0, 1))
      zz <- length(v1[1,  ])
      if(zz > 2)
         for(i in 2:zz) {
            v1[, 1] <- v1[, 1] + v1[, i]
            lines(c1, v1[, 1])
         }
   }
   invisible()
}
rpoly <- function(n, cov, fit)
{
   if(n < 1)
      stop("n is wrong")
   if(is.matrix(cov) == F)
      cov <- matrix(cov, nrow = 1)
   if(length(cov[1,  ]) != fit$ncov)
      stop("wrong number of covariates")
   if(n > 1 && length(cov[, 1]) == 1)
      cov <- matrix(cov, nrow = n, ncol = fit$ncov, byrow = T)
   if(n != length(cov[, 1]))
      stop("cov has wrong number of rows")
   vv <- ppoly(cov, fit)
   ww <- runif(n)
   zz <- rep(fit$nclass, n)
   for(i in 2:fit$nclass)
      vv[, i] <- vv[, i] + vv[, (i - 1)]
   for(i in fit$nclass:1)
      zz[ww < vv[, i]] <- i
   if(min(fit$classnames) == 0)
      zz <- zz - 1
   zz
}
poly.summary <- function(fit)
{
   it <- fit$method
   cat("========================POLYCLASS summary=======================\n")
   cat(paste("The fit was obtained with\n  "))
   cat("\b\b")
   print(fit$call)
   cat(paste("There were",fit$nclass,"classes and", fit$ncov,"covariates.\n"))
   if(it == 1) cat(paste("There were", fit$sample, "trial cases and",
      fit$tsample, "test cases.\n\n"))
   else cat(paste("There were",fit$sample,"cases.\n\n"))
   if(it == 0) {
      cat("The model selection was carried out using AIC.\n")
      if(0.99 < fit$penalty/log(fit$wgtsum)&&fit$penalty/log(fit$wgtsum) <1.01){
         cat(paste("The penalty was the default, log("))
         cat(paste(round(fit$wgtsum), "\b)="))
         cat(paste(round(log(fit$wgtsum), 2), "\b.\n"))
      }
      else{
          cat(paste("The penalty was", round(fit$penalty, 2), 
            "\b, the default would have been log("))
          cat(paste("\b",round(fit$wgtsum), "\b)="))
          cat(paste(round(log(fit$wgtsum), 2), "\b.\n"))
      }
   }
   if(it == 1) cat("The model selection was carried out using a test set.\n")
   if(it == 2) cat(paste("The model selection was carried out using", fit$cv,
      "\b-fold cross-validation.\n"))
   if((it == 1 || it == 2) && fit$select==0) {
      a2 <- range(fit$loss + diag(rep(1, fit$nclass)))
      if(a2[2] == 1 && a2[1] == 1) cat("The standard loss-matrix was used.\n")
      else cat("A loss matrix was provided.\n")
   }
   if((it == 1 || it == 2) && fit$select ==2) {
      cat("The sum of squared probabilities was used for the loss.\n")
   }
   if((it == 1 || it == 2) && fit$select ==1) {
      cat("Minus the test set log likelihood was used for the loss.\n")
   }
   if(length(fit$logl)<12)fit$logl<-t(as.matrix(fit$logl))
   a2 <- fit$logl[fit$logl[,1]==fit$ndim,  ]
   cat(paste("The model had dimension", fit$ndim, 
       "\b, log-likelihood",round(a2[3], 2)))
   if(it == 0) cat(paste(" and AIC", round( a2[2], 2), "\b.\n\n"))
   if(it == 1) cat(paste(" and loss", round( a2[2], 2), "\b.\n\n"))
   if(it == 2) cat(paste(" and AIC", round( a2[2], 2), "\b.\n\n"))
   if(it == 2){
      cat(paste("The penalty was cross-validated between", 
         round(fit$cv.aic[1],2)))
      if(fit$cv.aic[2]<0) cat(paste(" and Inf to",round(fit$cv.aic[4],2)))
      else cat(paste(" and",round(fit$cv.aic[2],2),"to", 
        round(fit$cv.aic[4],2)))
      cat(paste(" (loss",round(fit$cv.aic[3],2),"\b).\n"))
         cat(paste("The default penalty would have been log(",
         round(fit$wgtsum), "\b)=", round(log(fit$wgtsum), 2), "\b.\n"))
   }
   cat("The locations of the knots:\n")
   dimnames(fit$knots)[[2]][1] <- "Number"
   print(round(fit$knots, 3))
   cat(paste("\n  There are", fit$nbas, "basis functions, summarized below:\n"))
   a3 <- length(dimnames(fit$fcts)[[2]])
   for(i in 5:a3)
      dimnames(fit$fcts)[[2]][i] <- paste("Class",dimnames(fit$fcts)[[2]][i])
   print(round(fit$fcts, 3))
   cat("The first basis function is the constant function. For all others,\n")
   cat("the first column and the third column indicate on which covariates\n")
   cat("that basis function depends. If the third column is NA, the basis\n")
   cat("function depends on only one covariate.\n")
   cat("For the nonconstant basis functions the second and the fourth column\n")
   cat("indicate on which knot the function depend. If these columns are NA,\n")
   cat("the basis function is linear in this covariate.\n")
   cat("The remaining columns give the coefficients.\n")
   cat("\n")
   cat("================================================================\n")
   if(fit$method==0)
   cat("The influence of the penalty parameter is summarized below:\n")
   if(fit$method==1)
   cat("The effect of the penalty in the final run is summarized below:\n")
   if(fit$method==2)
   cat("The equivalence of the penalty parameter is summarized below:\n")
   dimnames(fit$logl)[[1]] <- rep("",length(fit$logl[,1]))
   fit$logl[,3:5]_fit$logl[,3:5]/fit$wgtsum
   if(fit$method==1)fit$logl[,6:8]_fit$logl[,6:8]/fit$twgtsum
   fit$logl[,"AIC"]_-fit$logl[,"AIC"]
   print(round(fit$logl, 3))
   fit$logl[,"AIC"]_-fit$logl[,"AIC"]
   if(fit$method==2){
  cat("The relation between the CV-loss and the penalty is summarized below:\n")
   dimnames(fit$cv.tab)[[1]] <- rep("",length(fit$cv.tab[,1]))
   print(round(fit$cv.tab, 3))}
   cat("================================================================\n")
   cat("The importance-anova decomposition is:\n")
   anova <- fit$anova
   anova[,3] <- anova[,3]*100
   dimnames(anova) <- list(rep("",length(anova[,1])),
        c("Cov-1","Cov-2","Percentage"))
   print(round(anova,2))
   cat("================================================================\n")
   invisible()
}
poly.beta <- function(fit, which, xsp = 0.4, cex)
{
 plot(c(0, 1), c(0, 1), axes = F, xlab = "", xlim = c(0.1, 0.9), ylim =
                c(0.1, 0.9), ylab = "", type = "n", xaxs = "d", yaxs = "d")
        lines(c(0, 1, 1, 0, 0), c(1, 1, 0, 0, 1))
        xsp <- xsp/4
        if(missing(which))
                which <- fit$classnames
        if(fit$classnames[1] == 0)
                which <- which + 1
        if(missing(cex))
                cex <- par()$cex
        nb <- fit$nbas
        lines( c(4 * xsp,4*xsp),c(0,1))
   b11 <- fit$beta
   for(i in 1:nb) {
      b1 <- fit$fcts[i,  ]
      y1 <- 1 - (i - 0.5)/nb
      y0 <- 1 - (i - 1)/nb
      y2 <- 1 - (i - 0)/nb
      lines(c(0, 4 * xsp), c(y0, y0))
      lines(c(2 * xsp, 2 * xsp), c(y0, y2))
      aa <- fit$covnames[b1[1]]
      if(is.na(b1[2])) aa <- paste(aa, "linear")
      else aa <- paste(aa, "at", signif(fit$knots[b1[1], b1[2] + 1], 2))
      if(i==1)aa <- "constant"
      text(xsp, y1, aa, cex = cex)
      if(is.na(b1[3]) == F) {
         aa <- fit$covnames[b1[3]]
         if(is.na(b1[4])) aa <- paste(aa, "linear")
         else aa <- paste(aa, "at", signif(fit$knots[b1[3], b1[4] + 1], 2))
         text(xsp * 3, y1, aa, cex = cex)
      }
      lines(c(4 * xsp + 0.03, 0.97), c(1 - (i - 0.1)/nb, 1 - (i - 0.1)/nb))
   }
   b2 <- range(b11)
   for(i in 1:nb) {
      b1 <- round(((0.92-4*xsp)*(b11[i,]-b2[1]))/(b2[2]-b2[1])+4*xsp+0.04,2)
      aa <- rep(1, length(b1))
      for(j in 1:length(b1)) aa[j] <- sum(abs(b1[1:j] - b1[j]) < 0.01)
      bb <- max(aa)
      if(bb > 1) bb <- 0.7/(nb * bb)
      for(j in which) text(b1[j], 1 - (i - 0.2)/nb + bb * (aa[j] - 1), 
         as.character(fit$classnames[j]))
   }
   invisible()
}
